/// tksmfmod.cpp
///
/// (c) 2002-2003 Bastian Spiegel <bs@tkscript.de>
/// TKS plugin distributed under terms of the GNU General Public License (GPL).
/// the original source has been modified to compile _correctly_ with visualc++ (++!).
/// the firelight routines excessively use inline (fpu) assembler so it might be some
/// kind of a challenge to port this beauty to something else than win32/visualc
/// --we gave up on it, now it's your try.
///
///
/// 290902 <bs> wrote interface, created files
/// 260603 <bs> removed from "host" to a plugin
///

/******************************************************************************/
/* MiniFMOD public source code release.                                       */
/* This source is provided as-is.  Firelight Multimedia will not support      */
/* or answer questions about the source provided.                             */
/* MiniFMOD Sourcecode is copyright (c) 2000, Firelight Multimedia.           */
/* MiniFMOD Sourcecode is in no way representative of FMOD 3 source.          */
/* Firelight Multimedia is a registered business name.                        */
/* This source must not be redistributed without this notice.                 */
/******************************************************************************/

#include <stdio.h>
#include <yac.h>
#include <yac_host.h>

extern YAC_Host *yac_host;

#include "tksmfmod.h"
#include "mfmod_minifmod.h"
#include "mfmod_sound.h"


// ---------------------------------------------------------------------------
YAC_Host        *yac_host;
TKSFMusicXM     *ctemplate;
YAC_CommandList *ctemplate_cmdlist;

void YAC_Init(YAC_Host *_host) {
	yac_host=_host;
	ctemplate=new TKSFMusicXM();
	ctemplate_cmdlist=ctemplate->yacGetCommandList();
	yac_host->yacRegisterClass(ctemplate, 1, ctemplate_cmdlist);
}
void YAC_Exit(YAC_Host *_host) {
	delete ctemplate; ctemplate=0;
	delete ctemplate_cmdlist; ctemplate_cmdlist=0;
}
// ---------------------------------------------------------------------------

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                   ~
//~ TKSFMusicXM                       ~
//~                                   ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
YAC_C(TKSFMusicXM, "FMusicXM");

TKSFMusicXM::TKSFMusicXM(void) {
	fmod=0;
	fmod_initialized=0;
	fmod_playing=0;
	volume=1.0f;
}

TKSFMusicXM::~TKSFMusicXM() {
	tksinter_free();
}

void TKSFMusicXM::tksinter_init(sS32 _freq, sS32 _numsamples) {
	yac_host->yacPrint("FMusicXM::init:\n");
	if(!fmod_initialized)
	{
		FSOUND_Init(_freq, _numsamples);
		fmod_initialized=1;
	}
	yac_host->yacPrint("FMusicXM::init: OK\n");
}

void TKSFMusicXM::tksinter_loadLocal(YAC_Object *_s, YAC_Value *_r) {
	if(fmod_initialized)
	{
		if(YAC_CHK(_s, YAC_CLID_STRING))
		{
			YAC_String *s=(YAC_String*)_s;
			YAC_Object *sio=(YAC_Object*)yac_host->yacNew("File");
			if(sio->yacStreamOpenLocal((sChar*)s->chars, 0))
			{
				// file opened, hand stream ptr over to fmod loader
				FMUSIC_MODULE *m=FMUSIC_LoadSong(0, sio);
				yac_host->yacDelete(sio);
				if(m)
				{
					fmod=(void*)m;
					YAC_RETI(1);
					return;
				}
			}
		}
	}
	YAC_RETI(0);
}

void TKSFMusicXM::tksinter_load(YAC_Object *_s, YAC_Value *_r) {
	if(fmod_initialized)
	{
		if(YAC_CHK(_s, YAC_CLID_STRING))
		{
			YAC_String *s=(YAC_String*)_s;
			YAC_Object *sio=(YAC_Object*)yac_host->yacNew("File");
			if(sio->yacStreamOpenLogic((sChar*)s->chars))
			{
				// file opened, hand stream ptr over to fmod loader
				FMUSIC_MODULE *m=FMUSIC_LoadSong(0, sio);
				yac_host->yacDelete(sio);
				if(m)
				{
					fmod=(void*)m;
					YAC_RETI(1);
					return;
				}
			}
		}
	}
	YAC_RETI(0);
}

void TKSFMusicXM::tksinter_play(void) {
	if(fmod)
	{
		FMUSIC_PlaySong((FMUSIC_MODULE *)fmod);
		fmod_playing=1;
	}
}

void TKSFMusicXM::tksinter_stop(void) {
	if(fmod)
	{
		fmod_playing=0;
		FMUSIC_StopSong((FMUSIC_MODULE *)fmod);
	}
}

void TKSFMusicXM::tksinter_getOrder(YAC_Value *_r) {
	if(fmod)
	{
		YAC_RETI(FMUSIC_GetOrder((FMUSIC_MODULE *)fmod));
		return;
	}
	YAC_RETI(0);
}

void TKSFMusicXM::tksinter_getRow(YAC_Value *_r) {
	if(fmod)
	{
		YAC_RETI(FMUSIC_GetRow((FMUSIC_MODULE *)fmod));
		return;
	}
	YAC_RETI(0);
}

void TKSFMusicXM::tksinter_getTime(YAC_Value *_r) {
	if(fmod)
	{
		YAC_RETI(FMUSIC_GetTime((FMUSIC_MODULE *)fmod));
		return;
	}
	YAC_RETI(0);
}

void TKSFMusicXM::tksinter_free(void) {
	if(fmod)
	{
		// ---- free FMUSIC structures ----
		FMUSIC_FreeSong((FMUSIC_MODULE*)fmod);
		fmod=0;
		fmod_initialized=0;
		fmod_playing=0;
	}
}

void TKSFMusicXM::tksinter_pause(sS32 _p) {
	fmod_playing=(fmod_initialized&&_p);
}

void TKSFMusicXM::tksinter_setVolume(sF32 _v) {
	volume=_v;
}

void TKSFMusicXM::tksinter_getVolume(YAC_Value *_r) {
	YAC_RETF(volume);
}

void TKSFMusicXM::tksinter_render(YAC_Object *_fa) {
	if(YAC_CHK(_fa, YAC_CLID_FLOATARRAY))
	{
		sF32 *d=(sF32 *)_fa->yacArrayGetPointer();
		sS32 rs=(sS32)_fa->yacArrayGetWidth()>>1; // STEREO samples
		if(rs!=(FSOUND_BlockSize>>2)) // STEREO samples, 2 bytes/ sample
		{
			// XXX assumes constant requested block size!
			if(fmod_playing)
			{
				FSOUND_Software_Fill(d, rs);
				// ---- scale replay output, floats are in -32767..32767 range, need -1, 1 ----
				sF32 *s=d;
				sF32 sv=(1.0f/32767.0f)*volume;
				sSI i;
				for(i=0; i<rs; i++)
				{
					*s++=(*s)*sv;
					*s++=(*s)*sv;
				}
			}
		}
		else
		{
			::fprintf(stderr, "FMusicXM::beginBlock: requested #samples (%i) does not match FSOUND_BlockSize (%i).\n",
				rs, FSOUND_BlockSize);
		}
	}
}


